from functools import wraps, partial
from again.utils import unique_hex
from ..utils.stats import Stats, Aggregator
from ..exceptions import VykedServiceException

import asyncio
import logging
import socket
import setproctitle
import time


def publish(func):
    """
    publish the return value of this function as a message from this endpoint
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):  # outgoing
        payload = func(self, *args, **kwargs)
        payload.pop('self', None)
        self._publish(func.__name__, payload)
        return None

    wrapper.is_publish = True

    return wrapper


def subscribe(func):
    """
    use to listen for publications from a specific endpoint of a service,
    this method receives a publication from a remote service
    """
    wrapper = _get_subscribe_decorator(func)
    wrapper.is_subscribe = True
    return wrapper


def xsubscribe(func=None, strategy='DESIGNATION'):
    """
    Used to listen for publications from a specific endpoint of a service. If multiple instances
    subscribe to an endpoint, only one of them receives the event. And the publish event is retried till
    an acknowledgment is received from the other end.
    :param func: the function to decorate with. The name of the function is the event subscribers will subscribe to.
    :param strategy: The strategy of delivery. Can be 'RANDOM' or 'LEADER'. If 'RANDOM', then the event will be randomly
    passed to any one of the interested parties. If 'LEADER' then it is passed to the first instance alive
    which registered for that endpoint.
    """
    if func is None:
        return partial(xsubscribe, strategy=strategy)
    else:
        wrapper = _get_subscribe_decorator(func)
        wrapper.is_xsubscribe = True
        wrapper.strategy = strategy
        return wrapper


def _get_subscribe_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        coroutine_func = func
        if not asyncio.iscoroutine(func):
            coroutine_func = asyncio.coroutine(func)
        return (yield from coroutine_func(*args, **kwargs))

    return wrapper


def request(func):
    """
    use to request an api call from a specific endpoint
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        params = func(self, *args, **kwargs)
        self = params.pop('self', None)
        entity = params.pop('entity', None)
        app_name = params.pop('app_name', None)
        request_id = unique_hex()
        params['request_id'] = request_id
        future = self._send_request(app_name, endpoint=func.__name__, entity=entity, params=params)
        return future

    wrapper.is_request = True
    return wrapper


def api(func):  # incoming
    """
    provide a request/response api
    receives any requests here and return value is the response
    all functions must have the following signature
        - request_id
        - entity (partition/routing key)
        followed by kwargs
    """
    wrapper = _get_api_decorator(func)
    return wrapper


def deprecated(func=None, replacement_api=None):
    if func is None:
        return partial(deprecated, replacement_api=replacement_api)
    else:
        wrapper = _get_api_decorator(func=func, old_api=func.__name__, replacement_api=replacement_api)
        return wrapper


def _get_api_decorator(func=None, old_api=None, replacement_api=None):
    @asyncio.coroutine
    @wraps(func)
    def wrapper(*args, **kwargs):
        _logger = logging.getLogger(__name__)
        start_time = int(time.time() * 1000)
        self = args[0]
        rid = kwargs.pop('request_id')
        entity = kwargs.pop('entity')
        from_id = kwargs.pop('from_id')
        wrapped_func = func
        result = None
        error = None
        failed = False

        status = 'succesful'
        success = True
        if not asyncio.iscoroutine(func):
            wrapped_func = asyncio.coroutine(func)

        Stats.tcp_stats['total_requests'] += 1

        try:
            result = yield from asyncio.wait_for(asyncio.shield(wrapped_func(self, **kwargs)), 60*10)

        except asyncio.TimeoutError as e:
            Stats.tcp_stats['timedout'] += 1
            error = str(e)
            status = 'timeout'
            success = False
            failed = True
            logging.exception("TCP request had a timeout for method %s", func.__name__)

        except VykedServiceException as e:
            Stats.tcp_stats['total_responses'] += 1
            error = str(e)
            status = 'handled_error'
            _logger.error('Handled exception %s for method %s ', e.__class__.__name__, func.__name__)

        except Exception as e:
            Stats.tcp_stats['total_errors'] += 1
            error = str(e)
            status = 'unhandled_error'
            success = False
            failed = True
            _logger.exception('Unhandled exception %s for method %s ', e.__class__.__name__, func.__name__)

        else:
            Stats.tcp_stats['total_responses'] += 1

        end_time = int(time.time() * 1000)

        hostname = socket.gethostname()
        service_name = '_'.join(setproctitle.getproctitle().split('_')[:-1])

        logd = {
            'endpoint': func.__name__,
            'time_taken': end_time - start_time,
            'hostname': hostname, 'service_name': service_name
        }
        logging.getLogger('stats').debug(logd)
        _logger.debug('Time taken for %s is %d milliseconds', func.__name__, end_time - start_time)

        # call to update aggregator, designed to replace the stats module.
        Aggregator.update_stats(endpoint=func.__name__, status=status, success=success,
                                server_type='tcp', time_taken=end_time - start_time)

        if not old_api:
            return self._make_response_packet(request_id=rid, from_id=from_id, entity=entity, result=result,
                                              error=error, failed=failed)
        else:
            return self._make_response_packet(request_id=rid, from_id=from_id, entity=entity, result=result,
                                              error=error, failed=failed, old_api=old_api,
                                              replacement_api=replacement_api)

    wrapper.is_api = True
    return wrapper



import logging
import asyncio
import setproctitle
from collections import defaultdict
import socket


class Stats:
    hostname = socket.gethostbyname(socket.gethostname())
    service_name = '_'.join(setproctitle.getproctitle().split('_')[:-1])
    # hostd = {'hostname': '', 'service_name': ''}
    http_stats = {'total_requests': 0, 'total_responses': 0, 'timedout': 0, 'total_errors': 0}
    tcp_stats = {'total_requests': 0, 'total_responses': 0, 'timedout': 0, 'total_errors': 0}

    @classmethod
    def periodic_stats_logger(cls):
        logd = defaultdict(lambda: 0)
        logd['hostname'] = cls.hostname
        logd['service_name'] = cls.service_name

        for key, value in cls.http_stats.items():
            logd[key] += value
            logd['http_' + key] = value

        for key, value in cls.tcp_stats.items():
            logd[key] += value
            logd['tcp_' + key] = value

        _logger = logging.getLogger('stats')
        _logger.info(dict(logd))

        asyncio.get_event_loop().call_later(120, cls.periodic_stats_logger)


class StatUnit:
    MAXSIZE = 10

    def __init__(self, key=None):
        self.key = key
        self.average = 0
        self.values = list()
        self.count = 0
        self.success_count = 0
        self.sub = dict()

    def update(self, val, success):
        self.values.append(val)
        if len(self.values) > self.MAXSIZE:
            self.values.pop(0)

        self.average = sum(self.values) / len(self.values)
        self.count += 1
        if success:
            self.success_count += 1

    def to_dict(self):

        d = dict({'count': self.count, 'average': self.average, 'success_count': self.success_count, 'sub': dict()})
        for k, v in self.sub.items():
            d['sub'][k] = v.to_dict()
        return d

    def __str__(self):
        return "{} {} {} {}".format(self.key, self.sum, self.average, self.count)


class Aggregator:
    _stats = StatUnit(key='total')

    @classmethod
    def recursive_update(cls, d, new_val, keys, success):
        if len(keys) == 0:
            return

        try:
            key = keys.pop()
            value = d[key]

        except KeyError:
            value = StatUnit(key=key)
            d[key] = value

        finally:
            value.update(new_val, success)
            cls.recursive_update(value.sub, new_val, keys, success)

    @classmethod
    def update_stats(cls, endpoint, status, time_taken, server_type, success=True):

        cls._stats.update(val=time_taken, success=success)
        cls.recursive_update(cls._stats.sub, time_taken, keys=[status, endpoint, server_type], success=success)

    @classmethod
    def dump_stats(cls):
        return cls._stats.to_dict()

    @classmethod
    def periodic_aggregated_stats_logger(cls):
        hostname = socket.gethostname()
        service_name = '_'.join(setproctitle.getproctitle().split('_')[:-1])

        logd = cls._stats.to_dict()
        logs = []
        for server_type in ['http', 'tcp']:
            try:
                server_type_d = logd['sub'][server_type]['sub']
            except KeyError:
                continue
            for k, v in server_type_d.items():
                d = dict({
                    'method': k,
                    'server_type': server_type,
                    'hostname': hostname,
                    'service_name': service_name,
                    'average_response_time': v['average'],
                    'total_request_count': v['count'],
                    'success_count': v['success_count']
                })
                for k2, v2 in v['sub'].items():
                    d['CODE_{}'.format(k2)] = v2['count']
                logs.append(d)

        _logger = logging.getLogger('stats')
        for logd in logs:
            _logger.info(dict(logd))

        asyncio.get_event_loop().call_later(300, cls.periodic_aggregated_stats_logger)



__all__ = ['Host', 'TCPServiceClient', 'TCPService', 'HTTPService', 'HTTPServiceClient', 'api', 'request', 'subscribe',
           'publish', 'xsubscribe', 'get', 'post', 'head', 'put', 'patch', 'delete', 'options', 'trace',
           'Registry', 'RequestException', 'Response', 'Request', 'log', 'setup_logging',
           'deprecated', 'VykedServiceException', 'VykedServiceError', '__version__']

from .host import Host  # noqa
from .services import (TCPService, HTTPService, HTTPServiceClient, TCPServiceClient)  # noqa
from .decorators.http import (get, post, head, put, patch, delete, options, trace)  # noqa
from .decorators.tcp import (api, request, subscribe, publish, xsubscribe, deprecated)  # noqa
from .registry import Registry  # noqa
from .utils import log  # noqa
from .exceptions import RequestException, VykedServiceError, VykedServiceException  # noqa
from .utils.log import setup_logging  # noqa
from .wrappers import Response, Request  # noqa

__version__ = '2.1.91'
